#include <ngx_http.h>

static ngx_str_t TEXT_PLAIN = ngx_string("text/plain");

ngx_shm_zone_t *ngx_shm_fragmentation_shm_zone = NULL;

static ngx_str_t ngx_shm_fragmentation_shm_name = ngx_string("shm_fragmentation_module");

static ngx_int_t ngx_shm_fragmentation_handler(ngx_http_request_t *r);

static char *ngx_shm_fragmentation(ngx_conf_t *cf, ngx_command_t *cmd, void *conf);

static ngx_int_t ngx_shm_fragmentation_init_shm_zone(ngx_shm_zone_t *shm_zone, void *data);

static ngx_int_t ngx_shm_fragmentation_postconfig(ngx_conf_t *cf);

static size_t SHM_SIZE = 512 * 4096;

static ngx_command_t  ngx_shm_fragmentation_commands[] = {
    { ngx_string("shm_fragmentation"),
      NGX_HTTP_LOC_CONF|NGX_CONF_NOARGS,
      ngx_shm_fragmentation,
      NGX_HTTP_LOC_CONF_OFFSET,
      0,
      NULL },
    ngx_null_command
};

static ngx_http_module_t  ngx_shm_fragmentation_module_ctx = {
    NULL,                                   /* preconfiguration */
    ngx_shm_fragmentation_postconfig,       /* postconfiguration */

    NULL,                                   /* create main configuration */
    NULL,                                   /* init main configuration */

    NULL,                                   /* create server configuration */
    NULL,                                   /* merge server configuration */

    NULL,                                   /* create location configuration */
    NULL                                    /* merge location configuration */
};

ngx_module_t  ngx_shm_fragmentation_module = {
    NGX_MODULE_V1,
    &ngx_shm_fragmentation_module_ctx,     /* module context */
    ngx_shm_fragmentation_commands,        /* module directives */
    NGX_HTTP_MODULE,                       /* module type */
    NULL,                                  /* init master */
    NULL,                                  /* init module */
    NULL,                                  /* init process */
    NULL,                                  /* init thread */
    NULL,                                  /* exit thread */
    NULL,                                  /* exit process */
    NULL,                                  /* exit master */
    NGX_MODULE_V1_PADDING
};


static ngx_int_t
ngx_shm_fragmentation_postconfig(ngx_conf_t *cf)
{
	ngx_shm_fragmentation_shm_zone = ngx_shared_memory_add(cf, &ngx_shm_fragmentation_shm_name, SHM_SIZE, &ngx_shm_fragmentation_module);

	if (ngx_shm_fragmentation_shm_zone == NULL) {
		return NGX_ERROR;
	}

	ngx_shm_fragmentation_shm_zone->init = ngx_shm_fragmentation_init_shm_zone;
	ngx_shm_fragmentation_shm_zone->data = (void *) 1;


    return NGX_OK;
}

static ngx_int_t
ngx_shm_fragmentation_init_shm_zone(ngx_shm_zone_t *shm_zone, void *data)
{
    if (data) {
        shm_zone->data = data;
        return NGX_OK;
    }

    shm_zone->data = shm_zone;

    return NGX_OK;
}


static char *
ngx_shm_fragmentation(ngx_conf_t *cf, ngx_command_t *cmd, void *conf)
{
    ngx_http_core_loc_conf_t             *clcf = ngx_http_conf_get_module_loc_conf(cf, ngx_http_core_module);

    clcf->handler = ngx_shm_fragmentation_handler;

    return NGX_CONF_OK;
}


static ngx_str_t ERROR = ngx_string("Error");
static ngx_str_t SUCCESS = ngx_string("OK");


static ngx_int_t
ngx_shm_fragmentation_handler(ngx_http_request_t *r)
{
	ngx_buf_t     *b;
	ngx_chain_t   out;
	void          *p;
	ngx_str_t     *text = NULL;
	ngx_str_t      value;
	size_t         i = 0;

	ngx_int_t rc;

	ngx_slab_pool_t *shpool = (ngx_slab_pool_t *) ngx_shm_fragmentation_shm_zone->shm.addr;

	ngx_http_arg(r, (u_char *) "x", 1, &value);
	if (value.len > 0) {
		i = ngx_atoi(value.data, value.len);
	}

	ngx_shmtx_lock(&shpool->mutex);
	// Using the size from x argument
//	if ((p = ngx_slab_alloc_locked(shpool, i)) == NULL) {
//		text = &ERROR;
//		ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0, "failed on %d of %d", i, SHM_SIZE);
//	} else {
//		text = &SUCCESS;
//		ngx_slab_free_locked(shpool, p);
//	}

	// Using a loop until fail for 250 times, restarting size when fail
	size_t j= 0;
	for (i = 0; i < SHM_SIZE; i+= 1024) {
		if ((p = ngx_slab_alloc_locked(shpool, i)) == NULL) {
			text = &ERROR;
			ngx_log_error(NGX_LOG_ERR, ngx_cycle->log, 0, "failed on %d of %d", i, SHM_SIZE);
			if (j < 250) {
				i = 0;
				j++;
			} else {
				break;
			}
		} else {
			text = &SUCCESS;
			ngx_slab_free_locked(shpool, p);
		}
	}

	ngx_shmtx_unlock(&shpool->mutex);

	ngx_http_discard_request_body(r);

	r->headers_out.status = NGX_HTTP_OK;
	r->headers_out.content_length_n = text->len;

	r->headers_out.content_type.data = TEXT_PLAIN.data;
	r->headers_out.content_type.len = TEXT_PLAIN.len;
	r->headers_out.content_type_len = TEXT_PLAIN.len;

	rc = ngx_http_send_header(r);

	if (rc == NGX_ERROR || rc > NGX_OK || r->header_only) {
		return rc;
	}

	b = ngx_create_temp_buf(r->pool, text->len);
	if (b == NULL) {
		return NGX_HTTP_INTERNAL_SERVER_ERROR;
	}

	b->last = ngx_copy(b->pos, text->data, text->len);
	b->memory = 1;
	b->last_buf = 1;
	b->last_in_chain = 1;
	b->flush = 1;

	out.buf = b;
	out.next = NULL;

	return ngx_http_output_filter(r, &out);
}
